from pathlib import Path

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from starlette.responses import FileResponse

from .config import Config
from .router import router
from .store import get_store

static_path = Path(__file__).parent.parent / "web/dist"


def create_app(config: Config):
    app = FastAPI(
        title="Debouncer",
        description="A proxy that debounce requests.",
        version="0.1.0",
    )

    app.state.config = config
    app.state.store = get_store(config.store_path)

    app.include_router(router)
    app.mount("/assets", StaticFiles(directory=static_path / "assets"), name="assets")

    @app.get("/")
    async def index():
        return FileResponse(static_path / "index.html")

    @app.on_event("shutdown")
    async def shutdown():
        app.state.store.close()

    return app



import asyncio

from httpx import Client, Request
from loguru import logger

from .schema import CallStatus, Endpoint
from .store import Store

client = Client()


async def dispatch(store: Store, endpoint: Endpoint):
    endpoint = store.get(endpoint.uid)
    if endpoint.call is None:
        raise ValueError(f"{endpoint.uid}: dispatching endpoint without call")

    # Make endpoint call
    logger.debug(f"{endpoint.uid}: dispatching call")
    request = Request(endpoint.method, endpoint.url)
    response = client.send(request)
    response.raise_for_status()

    endpoint.call.status = CallStatus.DISPATCHED
    store.save(endpoint.uid, endpoint)

    # Wait for eventual timeout endpoint
    logger.info(f"{endpoint.uid}: handling timeout")
    if endpoint.timeout > 0:
        logger.info(f"{endpoint.uid}: waiting {endpoint.timeout} seconds timeout")
        await asyncio.sleep(endpoint.timeout)

    # Close call or redispatch
    endpoint = store.get(endpoint.uid)
    await close_call_or_redispatch(store, endpoint)


async def close_call_or_redispatch(store: Store, endpoint: Endpoint) -> Endpoint:
    if endpoint.call is not None and endpoint.call.redispatch:
        logger.info(f"{endpoint.uid}: redispatching call")
        endpoint.call.redispatch = False
        store.save(endpoint.uid, endpoint)
        await dispatch(store, endpoint)
    else:
        logger.info(f"{endpoint.uid}: closing call")
        endpoint.call = None
        store.save(endpoint.uid, endpoint)

    return endpoint



from functools import lru_cache
from os import PathLike
from typing import List, cast
from uuid import UUID

from sqlitedict import SqliteDict  # type: ignore

from .schema import Endpoint


class Store:
    def __init__(self, path: PathLike):
        self.store = SqliteDict(path, autocommit=True)

    def close(self):
        self.store.close()

    def all(self) -> List[Endpoint]:
        values = cast(List[Endpoint], list(self.store.values()))
        return list(sorted(values, key=lambda x: x.url))

    def save(self, key: UUID, item: Endpoint):
        self.store[str(key)] = item

    def __contains__(self, key: UUID) -> bool:
        return str(key) in self.store

    def get(self, key: UUID) -> Endpoint:
        return cast(Endpoint, self.store[str(key)])

    def delete(self, key: UUID):
        del self.store[str(key)]


@lru_cache()
def get_store(store_path: PathLike):
    return Store(store_path)



from enum import Enum
from typing import Optional
from uuid import UUID, uuid5

from pydantic import BaseModel, HttpUrl

UUID_NAMESPACE = UUID("dabdee1d-f505-4b7c-8447-6b32df38eb77")


class CallStatus(str, Enum):
    CREATED = "created"
    DISPATCHED = "dispatched"
    LOCKED = "locked"


class Call(BaseModel):
    status: CallStatus = CallStatus.CREATED
    redispatch: bool = False


class EndpointMethod(str, Enum):
    POST = "POST"
    GET = "GET"


class BaseEndpoint(BaseModel):
    url: HttpUrl
    method: EndpointMethod = EndpointMethod.POST
    timeout: int = 30


class EndpointCreate(BaseEndpoint):
    pass


class Endpoint(BaseEndpoint):
    uid: UUID
    call: Optional[Call]

    @staticmethod
    def from_create(payload: EndpointCreate) -> "Endpoint":
        return Endpoint(uid=uuid5(UUID_NAMESPACE, payload.url), **dict(payload))



import secrets
from typing import List, Optional
from uuid import UUID

from fastapi import APIRouter, BackgroundTasks, Body, Depends, HTTPException, Query
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from loguru import logger

from .dispatch import close_call_or_redispatch, dispatch
from .schema import Call, CallStatus, Endpoint, EndpointCreate
from .state import State, get_state

http_basic = HTTPBasic(auto_error=False)


async def verify_auth(
    basic_credentials: HTTPBasicCredentials = Depends(http_basic),
    token: Optional[str] = Query(None),
    state: State = Depends(get_state),
):
    if state.config.credentials is None or len(state.config.credentials) == 0:
        return

    username_in = None
    password_in = None

    if basic_credentials is not None:
        username_in = basic_credentials.username
        password_in = basic_credentials.password

    if token is not None:
        token_credentials = token.split(":", maxsplit=1)
        if len(token_credentials) == 2:
            username_in = token_credentials[0]
            password_in = token_credentials[1]

    if not (username_in is None or password_in is None):
        password = state.config.credentials.get(username_in)
        if password is not None and secrets.compare_digest(password_in, password):
            return

    raise HTTPException(
        status_code=401,
        detail="Incorrect credentials",
        headers={"WWW-Authenticate": "Basic"},
    )


router = APIRouter(
    prefix="/api",
    dependencies=[Depends(verify_auth)],
)


@router.get("/", response_model=List[Endpoint], tags=["endpoints"])
async def get_endpoints(state: State = Depends(get_state)):
    return state.store.all()


@router.post("/", response_model=Endpoint, status_code=201, tags=["endpoints"])
async def create_endpoint(
    body: EndpointCreate = Body(...),
    state: State = Depends(get_state),
):
    endpoint = Endpoint.from_create(body)
    state.store.save(endpoint.uid, endpoint)
    return endpoint


async def endpoint_ctx(
    uid: UUID,
    state: State = Depends(get_state),
):
    if uid not in state.store:
        raise HTTPException(status_code=404, detail="Endpoint not found")
    return state.store.get(uid)


@router.delete("/{uid}", tags=["endpoints"])
async def delete_endpoint(
    endpoint: Endpoint = Depends(endpoint_ctx),
    state: State = Depends(get_state),
):
    state.store.delete(endpoint.uid)
    return


@router.post("/{uid}", response_model=Call, tags=["calls"])
async def create_call(
    worker: BackgroundTasks,
    endpoint: Endpoint = Depends(endpoint_ctx),
    state: State = Depends(get_state),
):
    if endpoint.call is None:
        logger.debug(f"{endpoint.uid}: creating new call")
        endpoint.call = Call()
        state.store.save(endpoint.uid, endpoint)
        worker.add_task(dispatch, state.store, endpoint)

    # Initial request not yet dispatched, ignoring
    # if endpoint.call.status == CallStatus.CREATED:

    # Initial request dispatched but new requests can be ignored, ignoring
    # if endpoint.call.status == CallStatus.DISPATCHED:

    # Initial request dispatched and locked, new request will be redispatched
    # once the initial request is deleted
    if endpoint.call.status == CallStatus.LOCKED:
        logger.debug(f"{endpoint.uid}: call is locked, it should be redispatched")
        endpoint.call.redispatch = True
        state.store.save(endpoint.uid, endpoint)

    return endpoint.call


@router.post("/{uid}/lock", tags=["calls"])
async def lock_call(
    endpoint: Endpoint = Depends(endpoint_ctx),
    state: State = Depends(get_state),
):
    if endpoint.call is None:
        raise HTTPException(status_code=404, detail="Call not found")

    logger.debug(f"{endpoint.uid}: locking call")
    endpoint.call.status = CallStatus.LOCKED
    state.store.save(endpoint.uid, endpoint)
    return endpoint.call


@router.post("/{uid}/close", tags=["calls"])
async def close_call(
    endpoint: Endpoint = Depends(endpoint_ctx),
    state: State = Depends(get_state),
):
    endpoint = await close_call_or_redispatch(state.store, endpoint)
    return endpoint.call



from fastapi import Request

from .config import Config
from .store import Store


# pylint: disable=too-few-public-methods
class State:
    config: Config
    store: Store


def get_state(req: Request) -> State:
    return req.app.state



from functools import lru_cache
from typing import Dict, Optional

from pydantic import BaseSettings, Field


class Config(BaseSettings):
    port: int = Field(4000, env="PORT")
    store_path: str = Field("debouncer.db", env="STORE_PATH")
    credentials: Optional[Dict[str, str]] = Field(None, env="CREDENTIALS")


@lru_cache()
def get_config():
    return Config()
